#include "calculation.h"
#include "variable.h"
#include "slthread.h"
#include <fstream>
#include <limits>

void Calculation :: naiveDivision (unsigned pp)
{
  p = pp;
  for (unsigned o = 0;  o < operators . size ();  o ++)
  { operators [o] -> getDimSizes ();
    operators [o] -> setOutput ();
    operators [o] -> naiveDivision (p);
    operators [o] -> calcDestSizes ();
  }
}


void Calculation :: minimize (unsigned pp)
{
  { unsigned seed;
#ifdef SLABBER_MPI
    if (MPI :: COMM_WORLD . Get_rank () == 0)
    { ifstream rs ("/dev/urandom");
      rs . read (reinterpret_cast <char *> (& seed), sizeof (seed));
    }
    MPI :: COMM_WORLD . Bcast (& seed, 1, MPI_UNSIGNED, 0);
#else
    ifstream rs ("/dev/urandom");
    rs . read (reinterpret_cast <char *> (& seed), sizeof (seed));
#endif
    srand (seed);
  }

  p = pp;
  for (unsigned o = 0;  o < operators . size ();  o ++)
    operators [o] -> randomInit (p);
  vector <unsigned> order (operators . size ());
  randomPermutation (order);
  bool betterFound = true;
  while (betterFound)
  { betterFound = false;
    for (unsigned i = 0;  i < operators . size ();  i ++)
      betterFound = betterFound || operators [order [i]] -> minimize ();
  }
}


double Calculation :: minimize (unsigned pp, unsigned numberOfTries)
{
  if (pp <= 1)  {naiveDivision (pp);  return 0;}
  double bestCost = numeric_limits <double> :: max ();
  for (unsigned t = 0;  t < numberOfTries;  t ++)
  { minimize (pp);
    double cost = calcCost ();
    if (cost < bestCost)
    { bestCost = cost;
      for (unsigned i = 0;  i < operators . size ();  i ++)
        operators [i] -> saveBestBest ();
    }
  }
  for (unsigned i = 0;  i < operators . size ();  i ++)
    operators [i] -> restoreBestBest ();
  return bestCost;
}


void Calculation :: setDivision (unsigned pp, string scheme)
{
  p = pp;
  unsigned pos = 0;
  for (unsigned o = 0;  o < operators . size ();  o ++)
  { operators [o] -> getDimSizes ();
    operators [o] -> setOutput ();
    operators [o] -> setDivision (p, scheme, pos);
    operators [o] -> calcDestSizes ();
  }
  bool changed = true;
  while (changed)
  { changed = false;
    for (unsigned o = 0;  o < operators . size ();  o ++)
      changed = changed || operators [o] -> minimizeResponsibility ();
  }
}

#ifdef SLABBER_SHM
vector <ThreadInfo> threadInfos;
#if __cplusplus >= 201103L
vector <thread> threads;
#else
vector <pthread_t> threads;
#endif
unsigned calculateBegin = 0;
unsigned calculateEnd = 0;
#endif


#if defined (SLABBER_SHM) && __cplusplus < 201103L
void * threadRun (void * arg)
{
  ThreadInfo * threadInfo = reinterpret_cast <ThreadInfo *> (arg);
  pthread_setspecific (threadInfoKey, threadInfo);
  threadInfo -> calculation -> singleThreadRun (threadInfo);
  return 0;
}
#endif

void Calculation :: initialize (unsigned numThreads = 0)
{
#ifdef SLABBER_SHM
  :: numThreads = numThreads;
#endif

  for (unsigned o = 0;  o < operators . size ();  o ++)
  { operators [o] -> getDimSizes ();
    operators [o] -> setOutput ();
    operators [o] -> allocateData ();
  }

  if (numThreads == 0)  return;

#ifdef SLABBER_SHM
  threadInfos . resize (numThreads);
  for (unsigned i = 0;  i < numThreads;  i ++)
    {threadInfos [i] . threadIndex = i;  threadInfos [i] . calculation = this;}

#if __cplusplus >= 201103L
  threadInfo = & threadInfos [0];

  threads . resize (numThreads - 1);
  for (unsigned i = 0;  i < numThreads - 1;  i ++)
    threads [i] = thread (& Calculation :: singleThreadRun, * this, & threadInfos [i + 1]);
#else
  pthread_key_create (& threadInfoKey, NULL);

  pthread_barrier_init (& threadBarrier, NULL, numThreads);

  threads . resize (numThreads - 1);
  for (unsigned i = 0;  i < numThreads - 1;  i ++)
    pthread_create (& threads [i], 0, :: threadRun, & threadInfos [i + 1]);

  pthread_setspecific (threadInfoKey, & threadInfos [0]);
#endif
#endif
}


void Calculation :: run (unsigned cbegin, unsigned cend)
{
  if (cend == 0)  cend = length ();

#ifndef SLABBER_SHM
  for (unsigned o = cbegin;  o < cend;  o ++)
  { operators [o] -> getInput ();
    operators [o] -> operate ();
    operators [o] -> setOutput ();
  }
#else
  threadBarrier1 ();
  calculateBegin = cbegin;  calculateEnd = cend;
  threadBarrier2 ();

  for (unsigned o = calculateBegin;  o < calculateEnd;  o ++)
  {
    operators [o] -> getInput ();
    threadBarrier1 ();
    operators [o] -> operate ();
    operators [o] -> setOutput ();
    threadBarrier2 ();
  }
#endif
}


void Calculation :: stop ()
{
#ifdef SLABBER_SHM
  threadBarrier1 ();
  calculateEnd = 0;
  threadBarrier2 ();

#if __cplusplus >= 201103L
  for (unsigned i = 0;  i < threads . size ();  i ++)
    threads [i] . join ();
#else
  void * r;
  for (unsigned i = 0;  i < threads . size ();  i ++)
    pthread_join (threads [i], & r);

  pthread_key_delete (threadInfoKey);
#endif

#endif
}


#ifdef SLABBER_SHM
void Calculation :: singleThreadRun (ThreadInfo * threadInfo)
{
#if __cplusplus >= 201103L
  :: threadInfo = threadInfo;
#endif
  for (;;)
  {
    threadBarrier1 ();
    threadBarrier2 ();

    if (calculateEnd == 0)  break;

    for (unsigned o = calculateBegin;  o < calculateEnd;  o ++)
    {
      //if (threadInfo -> threadIndex == 0)  operators [o] -> getInput ();
      threadBarrier1 ();
      //ReduceOperator * ro = dynamic_cast <ReduceOperator *> (operators [o]);
      //if (! ro)  operators [o] -> operate ();
      operators [o] -> operate ();
      //else  ro -> reduce ();
      //if (threadInfo . threadIndex == 0)  operators [o] -> setOutput ();
      //operators [o] -> setOutput ();
      threadBarrier2 ();
    }
  }
}
#endif

