#ifndef CALCULATION_H
#define CALCULATION_H

#include "operator.h"
#include "variable.h"
#include "slthread.h"


class Calculation
{
public:
  Calculation ()  {};

  void addOperator (Operator * op)  {operators . push_back (op);}

  Variable * newVariable (Operator * op, unsigned operand)
  { Variable * var = new Variable (op, operand, this);
    op -> plug (* var, operand);
    variables . push_back (var);
    return var;
  }

  Variable * newVariable (Variable & v, bool transposed)
  { Variable * var = new Variable (v, transposed);
    variables . push_back (var);
    return var;
  }

  void minimize (unsigned pp);
  double minimize (unsigned pp, unsigned numberOfTries);

  //void minimize (unsigned k, double cost);

  void naiveDivision (unsigned pp);

  void setDivision (unsigned pp, string scheme);

  void printDivision ()
  { for (unsigned i = 0;  i < operators . size ();  i ++)
      {operators [i] -> printDivision ();  cout << ' ';}
    cout << endl;
  }

  void printConnections ()
  { for (unsigned i = 0;  i < operators . size ();  i ++)
      operators [i] -> printConnections ();
  }

  double calcCost ()
  { double cost = 0.0;
    for (unsigned i = 0;  i < operators . size ();  i ++)
      cost += operators [i] -> calcCost (true);
    return cost;
  }

  void initialize (unsigned numThreads);

  void run (unsigned cbegin = 0, unsigned cend = 0);
  void stop ();

  unsigned length ()  {return operators . size ();}

private:
#ifdef SLABBER_SHM
  void singleThreadRun (ThreadInfo * threadInfo);
  friend void * threadRun (void *);
#endif

  vector <Operator *> operators;
  vector <Variable *> variables;
  unsigned p;
};


class Block
{
public:
  Block (Calculation & pcalculation)
  { calculation = & pcalculation;
    cbegin = calculation -> length ();
  }

  void end ()  {cend = calculation -> length ();}

  void run ()  {calculation -> run (cbegin, cend);}

private:
  Calculation * calculation;
  unsigned cbegin, cend;
};


inline Variable & trans (Variable & v)  {return * v . calculation -> newVariable (v, true);}


#endif

