// -*-C++-*-

#include <time.h>
#include <sys/time.h>
#include <string>

/// A time wrapper class
class GanTime
{
public:
  /// Create a zero time
  GanTime (void) { TimeValue = 0; }

  /// Create time by internal value (do not use this)
  GanTime (double tv) { TimeValue = tv; }

  /// return internal value (do not use this)
  double GetValue (void) { return TimeValue; }

  /// Subtract times
  GanTime operator- (GanTime othertime)
  { return GanTime (TimeValue - othertime.GetValue ()); }

  /// Add times
  GanTime operator+ (GanTime othertime)
  { return GanTime (TimeValue + othertime.GetValue ()); }

  /// does this time point into the past?
  bool Negative (void) const { return TimeValue < 0; }

  /// Get the hours of the time (can be >24)
  unsigned Hours (void) const
  { return (unsigned) (TimeValue / 3600000000.0); }

  /// Get the minutes (range 0 to 59)
  unsigned Mins (void) const
  { return ((unsigned) (TimeValue / 60000000.0)) % 60; }

  /// Get the seconds (range 0 to 59)
  unsigned Secs (void) const
  { return ((unsigned) (TimeValue / 1000000.0)) % 60; }

  /// Get the milliseconds (range 0 to 999)
  unsigned MilliSecs (void) const
  { return ((unsigned) (TimeValue / 1000.0)) % 1000; }

  /// Get the microseconds (range 0 to 999)
  unsigned MicroSecs (void) const
  { return ((unsigned) TimeValue) % 1000; }

  /// Get the time in hours
  double InHours (void) const
  { return TimeValue / 3600000000.0; }

  /// Get the time in minutes
  double InMins (void) const
  { return TimeValue / 60000000.0; }

  /// Get the time in seconds
  double InSecs (void) const
  { return TimeValue / 1000000.0; }

  /// Get the time in milliseconds
  double InMilliSecs (void) const
  { return TimeValue / 1000.0; }

  /// Get the time in microseconds
  double InMicroSecs (void) const
  { return TimeValue; }

  /// Print the time (format: ??h??m??s??ms??us)
  void Print (ostream &s = cout) const
  { if (Negative ()) s << '-';
    unsigned T;
    T = Hours ();
    if (T != 0) { s << T << 'h'; }
    T = Mins ();
    if (T != 0) { s << T << 'm'; }
    T = Secs ();
    if (T != 0) { s << T << 's'; }
    T = MilliSecs ();
    if (T != 0) { s << T << "ms"; }
    T = MicroSecs ();
    if (T != 0) { s << T << "us"; }
  }

private:
  /// the time in microseconds
  double TimeValue;
};

/// \file

/// Returns the time since 00:00 today
inline GanTime GanTimeOfDay (void)
{
    struct timeval tv;
    gettimeofday (&tv, 0);
    return GanTime ((double) tv.tv_sec * 1000000.0 + (double) tv.tv_usec);
}

/// Returns a time that corresponds to CPU time
inline GanTime GanCPUTime (void)
{
    assert (CLOCKS_PER_SEC == 1000000);
    return GanTime ((double) clock ());
}

/// Print the time (format: ??h??m??s??ms??us)
inline ostream &operator<<(ostream &s, GanTime const &T)
{ T.Print (s);
  return s;
}
