#include "slabber.h"
#include "matrix.h"
#include <iostream>
#include "gantime.h"

using namespace std;

int main (int argc, char * argv [])
{
  SlabberInit (argc, argv);
  unsigned m, n, k, repeat, numThreads, naive;
  if (argc != 7)
  {
#ifdef SLABBER_MPI
    if (MPI :: COMM_WORLD . Get_rank () == 0)
#endif
    cerr << "nmfmu <A-rows> <A-cols> <W-cols=H-rows> <#iterations> <#threads/processes> <optimized=1,naive=0>\n";
    SlabberFinalize ();
    exit (0);
  }
  m = atoi (argv [1]);
  n = atoi (argv [2]);
  k = atoi (argv [3]);
  repeat = atoi (argv [4]);
  numThreads = atoi (argv [5]);
  naive = atoi (argv [6]);

  Calculation cal;
  Block generateData (cal);
  Variable & a = generateMatrix (m, n, cal);
  Variable & w = generateMatrix (m, k, cal);
  Variable & h = generateMatrix (k, n, cal);
  generateData . end ();

  Block mainLoop (cal);
  h = pDiv (pMult (h, (trans (w) * a)), (trans (w) * w) * h);
  w = pDiv (pMult (w, (a * trans (h))), w * (h * trans (h)));
  mainLoop . end ();

  Block result (cal);
/*
  consumeMatrix (h);
  consumeMatrix (w);

  Variable & b = w * h;
  consumeMatrix (a);
  consumeMatrix (b);
*/
  result . end ();

  if (numThreads != 0 && ! naive)  cal . minimize (numThreads, 50);
  if (numThreads != 0 && naive)  cal . naiveDivision (numThreads);

#ifdef SLABBER_MPI
  MPI :: COMM_WORLD . Barrier ();
#endif
  GanTime t1 = GanTimeOfDay ();

  cal . initialize (numThreads);
  generateData . run ();
  for (unsigned i = 0;  i < repeat; i ++)
    mainLoop . run ();

  GanTime t2 = GanTimeOfDay ();

#ifdef SLABBER_MPI
  if (MPI :: COMM_WORLD . Get_rank () == 0)
#endif
  cout << m << ' ' << n << ' ' << k << ' ' << numThreads << ' ' << naive << ' ' << (t2 - t1) . InMilliSecs () / repeat << endl;

  result . run ();
  cal . stop ();

  SlabberFinalize ();
}

