#ifndef OPERATOR_H
#define OPERATOR_H

#include <iostream>
#include <vector>
#include <cstdlib>
#include <cassert>
#include "slthread.h"
#include "slab.h"
#include "permutation.h"
#include "variable.h"
#include "misc.h"
#include "region.h"

#define COMP_COST 0.01
#define COMM_COST 1.0

using namespace std;


class Calculation;
class Operator;


class Operand
{
public:
  Operand (bool pin)  {in = pin;}

  Operand (bool pin, unsigned dim)
    {in = pin;  dimensions . push_back (dim);  transposition . push_back (0);}

  Operand (bool pin, unsigned dim1, unsigned dim2)
  { in = pin;
    dimensions . push_back (dim1);  dimensions . push_back (dim2);
    transposition . push_back (0);  transposition . push_back (1);
  }

  Operand (bool pin, unsigned dim1, unsigned dim2, unsigned dim3)
  { in = pin;
    dimensions . push_back (dim1);
    dimensions . push_back (dim2);
    dimensions . push_back (dim3);
    transposition . push_back (0);
    transposition . push_back (1);
    transposition . push_back (2);
  }

  ~Operand ()  {if (! in)  delete data . getData ();}

  unsigned size (vector <unsigned> & dimSizes)
  { unsigned s = 1;
    for (unsigned i = 0;  i < dimensions . size ();  i ++)
      s *= dimSizes [dimensions [i]];
    return s;
  }

  void transpose ()
  { assert (dimensions . size () == 2);
    swap (transposition [0], transposition [1]);
  }

  void getInput ();

  void allocateData (vector <unsigned> & dimSizes);

  vector <unsigned> dimensions;
  bool in;
  Variable * variable;
  vector <unsigned> transposition;
  Slab data;
};


class Operator
{
public:

  //Operator ()  {divisionLocked = false;  divisionEnd = true;}
  Operator ()  {divisionEnd = true;}

  void plug (Variable & v, unsigned operand)
  {
    if (v . reference)
    { plug (* v . reference, operand);
      if (v . transposed)  operands [operand] . transpose ();
    }
    else
    { operands [operand] . variable = & v;
      if (v . sourceOperator -> operands . size () <= v . sourceOperand)
      { cerr << "error: Operand " << operand << " of " << label << " plugged from operand "
             << v . sourceOperand << " of " << v . sourceOperator -> label << ", which has only "
             << v . sourceOperator -> operands . size () << " operands.\n";
        exit (1);
      }
      if (operands [operand] . in)
      { v . destOperators . push_back (this);
        v . destOperands . push_back (operand);
      }
    }
  }

  void getDimSizes ();

  void naiveDivision (unsigned pp)
  { p = pp;
    unsigned maxDim = 0, maxDimSize = 0;
    for (unsigned i = 0;  i < dimensions;  i ++)  divisions [i] = 1;
    for (unsigned i = 0;  i < splitDimensions;  i ++)
      if (dimSizes [i] > maxDimSize)  {maxDimSize = dimSizes [i];  maxDim = i;}
  maxDim = 0;  // noch naiver !!!
    divisions [maxDim] = p;
    responsibility . resize (p);
    for (unsigned i = 0;  i < p;  i ++)  responsibility [i] = i;
  }

  bool minimize ();

  bool minimizeResponsibility ();

  void setDivision (unsigned pp, string & scheme, unsigned & pos);

  void beginDivision (unsigned pp)
  { p = pp;
    //if (! divisionLocked)
    { for (unsigned i = 0;  i < dimensions;  i ++)  divisions [i] = 1;
      //divisions [splitDimensions - 1] = p;
      divisions [0] = p;
    }
    //else getDivisions ();
    divisionEnd = false;
  }

# define BASIC_DIVISIONS // try only divisions of type 1,p,1,1

# ifdef BASIC_DIVISIONS

  void nextDivision ()
  {
    unsigned pPos = 0;
    while (divisions [pPos] != p)  pPos ++;
    if (pPos + 1 >= splitDimensions)  divisionEnd = true;
    divisions [pPos] = 1;
    divisions [pPos + 1] = p;
  }

# else

  void nextDivision ()
  { //if (divisions . size () == 1 || divisionLocked)  {divisionEnd = true;  return;}
    if (divisions . size () == 1)  {divisionEnd = true;  return;}
    unsigned pos = 0, d = 0, dp = 0;
    for (;;)
    { pos = 0;  d = 0;  dp = 0;
      do
      { dp = pos;  d = divisions [dp];
        divisions [pos] ++;
        for (unsigned i = 0;  i < pos;  i ++)  divisions [i] = 1;
        normalDivision ();
        pos ++;
      }
      while (d == divisions [dp] && pos < splitDimensions);
      bool restrictOk = true;
      for (unsigned i = 0;  i < splitDimensions;  i ++)
        if (divisions [i] != 1 && divisions [i] != p)  restrictOk = false;
      if (restrictOk)  break;
    }
    if (pos == splitDimensions && d == divisions [dp])  divisionEnd = true;
  }

# endif

  bool endDivision ()  {return divisionEnd;}

  void saveBest ()
    {bestDivisions = divisions;  bestResponsibility = responsibility;}

  void restoreBest ()
  { divisions = bestDivisions;  responsibility = bestResponsibility;
    calcDestSizes ();
  }

  void saveBestBest ()
    {bestBestDivisions = divisions;  bestBestResponsibility = responsibility;}

  void restoreBestBest ()
  {  divisions = bestBestDivisions;  responsibility = bestBestResponsibility;
     calcDestSizes ();
  }


  void printDivision (bool printResponsibility = false)
  { cout << label << ':';
    unsigned prod = 1;
    for (unsigned i = 0;  i < divisions . size ();  i ++)
    { if (i != 0)  cout << ',';
      cout << divisions [i];
      if (i < splitDimensions)  prod *= divisions [i];
    }
    //cout << "(" << prod << ")";
    if (printResponsibility)
    { cout << "[";
      for (unsigned i = 0;  i < responsibility . size ();  i ++)
      { if (i > 0)  cout << ',';
        cout << responsibility [i];
      }
      cout << "]";
    }
  }

  void printConnections ()
  { cout << label << ": ";
    for (unsigned i = 0;  i < operands . size ();  i ++)
      if (operands [i] . in)
        cout << operands [i] . variable -> sourceOperator -> label << '.'
             << operands [i] . variable -> sourceOperand << ' ';
    cout << endl;
  }

  virtual void calcDestSizes ()  {};

  unsigned dividedSize (unsigned d)
  { unsigned div = divisions [d];
    return (dimSizes [d] + div - 1) / div;
  }

  unsigned dividedOperandSize (unsigned o)
  { Operand & op = operands [o];
    unsigned s = 1;
    for (unsigned i = 0;  i < op . dimensions . size ();  i ++)
      s *= dividedSize (op . dimensions [i]);
    return s;
  }

  Region getRegion (unsigned o, unsigned p, bool transposed = false)
  { Operand & op = operands [o];
    vector <unsigned> sizes;  shuffle (op . dimensions, dimSizes, sizes);
    vector <unsigned> opDivs;  shuffle (op . dimensions, divisions, opDivs);
    vector <unsigned> pos;  index2pos (responsibility [p], opDivs, pos);
    if (transposed && ! op . transposition . empty ())
    { shuffle (op . transposition, opDivs);
      shuffle (op . transposition, pos);
      shuffle (op . transposition, sizes);
    }
    return Region (sizes) . divide (opDivs, pos);
  }

  double calcCommCost (unsigned o, unsigned pi)
  { Variable & v = * operands [o] . variable;
    if (operands [o] . in)
    { getDimSizes ();
      Region r = getRegion (o, pi, true);
      return   r . size ()
             - r . intersect (v . sourceOperator -> getRegion (v . sourceOperand, pi, false))  // true??
                 . size ();
    }
    else
    { double cost = 0.0;
      for (unsigned u = 0;  u < v . destOperators . size ();  u ++)
        cost += v . destOperators [u] -> calcCommCost (v . destOperands [u], pi);
      return cost;
    }
  }

  double calcCost (bool onlyIn = false);

  void randomInit (unsigned pp);

  unsigned operandCount ()  {return operands . size ();}

  Operand & getOperand (unsigned i)  {return operands [i];}

  void allocateData ();

  virtual void operate () = 0;

  void getInput ();

  void setOutput ()
  { for (unsigned o = 0;  o < operands . size ();  o ++)
      if (! operands [o] . in)
      { operands [o] . variable -> sourceOperator = this;
        operands [o] . variable -> sourceOperand = o;
      }
  }

protected:

  void setDimensions (unsigned d)
  { dimensions = d;
    dimSizes . resize (d);
    divisions . resize (d);  for (unsigned i = 0;  i < d;  i ++)  divisions [i] = 1;
  }

  //void getDivisions ();

  virtual double calcCompCost () = 0;

  Slab getOperandData (unsigned o, vector <int> * position = 0)
  {
#ifndef SLABBER_SHM
    if (position)
    { (* position) . resize (operands [o] . dimensions . size ());
      for (unsigned d = 0;  d < position -> size ();  d ++)
        (* position) [d] = 0;
    }
    return operands [o] . data;
#else
    unsigned threadIndex = getThreadInfo () . threadIndex;
    return operands [o] . data . slab
      (responsibility [threadIndex], operands [o] . dimensions, divisions, position);
#endif
  }

private:

  void normalDivision ()
  { unsigned prod = 1;
    for (unsigned i = 0;  i < splitDimensions;  i ++)  prod *= divisions [i];
    for (unsigned i = 0;  i < splitDimensions;  i ++)
    { prod /= divisions [i];
      divisions [i] = max (p / prod, 1u);
      prod *= divisions [i];
    }
  }

protected:
  unsigned dimensions;
  vector <unsigned> dimSizes;
  vector <Operand> operands;
  unsigned splitDimensions;
  unsigned p;
  //bool divisionLocked;
  vector <unsigned> divisions;
  vector <unsigned> responsibility;  // maps process index to slab index
  bool divisionEnd;
  vector <unsigned> bestDivisions;
  vector <unsigned> bestResponsibility;
  vector <unsigned> bestBestDivisions;
  vector <unsigned> bestBestResponsibility;
  const char *label;
};


class Pointwise : public Operator
{
public:
  Pointwise (unsigned dim, char const * plabel)
  { setDimensions (dim);
    operands . push_back (Operand (true, 0, 1));
    operands . push_back (Operand (true, 0, 1));
    operands . push_back (Operand (false, 0, 1));
    splitDimensions = dim;
    label = plabel;
  }

protected:
  double calcCompCost ()
  { double c = 0;
    for (unsigned i = 0;  i < dimensions;  i ++)  c *= dividedSize (i);
    return COMP_COST * c;
  }
};


class ReduceOperator : public Operator
{
public:
  friend class Operator;

  void reduce ();

private:
  void calcDestSizes ()
  { for (unsigned i = sourceDims;  i < dimensions;  i ++)  divisions [i] = 1;
    for (unsigned i = 0;  i < sourceDims;  i ++)
    { if (! reducedDims [i])  dimSizes [destDims [i]] = dimSizes [i];
      if (destDims [i] >= sourceDims)  divisions [destDims [i]] *= divisions [i];
    }
  }

protected:
  unsigned sourceDims;
  vector <unsigned> destDims;
  vector <bool> reducedDims;
};


inline Variable & Variable :: operator= (Variable & v)
{
  sourceOperator = v . sourceOperator;
  sourceOperand = v . sourceOperand;
  calculation = v . calculation;
  transposed = v . transposed;
  reference = v . reference;
  destOperators = v . destOperators;
  destOperands = v . destOperands;
  sourceOperator -> plug (* this, sourceOperand);
  return * this;
}


inline Variable :: Variable (Variable & v)
{
  sourceOperator = v . sourceOperator;
  sourceOperand = v . sourceOperand;
  calculation = v . calculation;
  transposed = v . transposed;
  reference = v . reference;
  destOperators = v . destOperators;
  destOperands = v . destOperands;
  sourceOperator -> plug (* this, sourceOperand);
}

inline void Variable :: print (ostream & o)
  {sourceOperator -> getOperand (sourceOperand) . data . print (o);}


#endif
