#include <iostream>
#include <vector>

class Region
{
public:
  Region (unsigned n)  {low . assign (n, 0);  high . assign (n, 0);}

  Region (std :: vector <unsigned> & dim)
    {high = dim;  low . assign (dim . size (), 0);}

  Region (std :: vector <unsigned> const & plow, std :: vector <unsigned> const & phigh)
    {low = plow;  high = phigh;}

  Region (Region const & r)  {low = r . low;  high = r . high;}

  Region & operator= (Region const & r)
    {low = r . low;  high = r . high;  return * this;}

  bool operator== (Region const & r) const
  { for (unsigned i = 0;  i < low . size ();  i ++)
      if (low [i] != r . low [i] || high [i] != r . high [i])  return false;
    return true;
  }

  Region intersect (Region const & b)
  { Region r (low . size ());
    for (unsigned i = 0;  i < low . size ();  i ++)
    { r . low [i] = std :: max (low [i], b . low [i]);
      r . high [i] = std :: min (high [i], b . high [i]);
      if (r . low [i] >= high [i])  return Region (low . size ());
    }
    return r;
  }

  Region divide (std :: vector <unsigned> const divisions, std :: vector <unsigned> const pos) const
  { Region r (low . size ());
    for (unsigned i = 0;  i < low . size ();  i ++)
    { unsigned dsize = (high [i] - low [i] + divisions [i] - 1) / divisions [i];
      r . low [i] = std :: min (low [i] + dsize * pos [i], high [i]);
      r . high [i] = std :: min (low [i] + dsize * (pos [i] + 1), high [i]);
    }
    return r;
  }

  Region operator- (std :: vector <unsigned> const & pos) const
  { Region r (low . size ());
    if (empty ())  return r;
    for (unsigned i = 0;  i < low . size ();  i ++)
    { r . low [i] = low [i] - pos [i];
      r . high [i] = high [i] - pos [i];
    }
    return r;
  }

  unsigned size () const
  { unsigned r = 1;
    for (unsigned i = 0;  i < low . size ();  i ++)
      r *= high [i] - low [i];
    return r;
  }

  bool empty () const
  { for (unsigned i = 0;  i < low . size ();  i ++)
      if (low [i] >= high [i])  return true;
    return false;
  }

  template <typename T>
  unsigned getSizes (vector <T> & sizes)
  { sizes . resize (low . size ());
    unsigned size = 1;
    for (unsigned i = 0;  i < low . size ();  i ++)
    { sizes [i] = high [i] - low [i];
      size *= sizes [i];
    }
    return size;
  }

  std :: vector <unsigned> const & getLow () const  {return low;}
  std :: vector <unsigned> const & getHigh () const  {return high;}

  void print (std :: ostream & s) const
  { s << '[';
    for (unsigned i = 0;  i < low . size ();  i ++)
    { if (i > 0)  s << ", ";
      s << low [i] << ':' << high [i];
    }
    s << ']';
  }

private:
  std :: vector <unsigned> low, high;
};


inline std :: ostream & operator<< (std :: ostream & s, Region const & r)  {r . print (s);  return s;}


#ifdef test_region
using namespace std;

int main ()
{
  unsigned d [] = {10, 11, 12, 13};
  vector <unsigned> dv (d, d + 4);
  cout << Region (dv) << endl
       << Region (dv) . divide (vector <unsigned> (4, 3), vector <unsigned> (4, 0)) << endl
       << Region (dv) . divide (vector <unsigned> (4, 3), vector <unsigned> (4, 1)) << endl
       << Region (dv) . divide (vector <unsigned> (4, 3), vector <unsigned> (4, 2)) << endl;
  unsigned l1 [] = {1, 2, 3, 4}, h1 [] = {8, 7, 6, 5}, l2 [] = {4, 4, 3, 3}, h2 [] = {7, 7, 5, 5};
  Region r1 (vector <unsigned> (l1, l1 + 4), vector <unsigned> (h1, h1 + 4)),
         r2 (vector <unsigned> (l2, l2 + 4), vector <unsigned> (h2, h2 + 4));
  cout << endl << r1 << endl << r2 << endl << r1 . intersect (r2) << endl;
}
#endif
